c*********************************************************************
c 
c File           nr.f
c Author         Jelena Ivanisevic
c Date           Octobar 16, 1996
c
c Description    Newton-Raphson's Method for zero finding
c
c Reference      Numerical Recepies 
c
c*********************************************************************
c      Function input is bracket x1, x2, and qubic coefficients
c      Function returns real root rtsafe and sets iflag
c         0 - exit on running rounoff error exit criteria
c         1 - exit on root(k) = root(k-1), exact root found
c        -2 - Nans or Inf in calculation - abort
c        -1 - Max number of iterations reached.
c********************************************************************
 
      real function rtsafe( x1, x2, iflag, a )
c        
         parameter (maxit = 1000)
         real a(0:3)
         integer iflag
         logical big
         real eps, poli,abs
         real x1, x2
         real xl, xh, f, df, e, temp, mepsd, outa, outb, dx, dxold
 
c        Calculate machine eps  * 0.5
         
         mepsd =.5 * eps()
*	 print *, 'mepsd', mepsd
c                 
         rtsafe = .5 * ( x1 + x2 )
         dxold = abs( x2 - x1 )
         dx = dxold
 
c        keep bracket xl, xh such that f(xl).lt.0 , f(xh).gt.0                   
 
         fl = poli(x1, a)
         fh = poli(x2, a)
 
c        Check for exceptions
 
         if( isNaN(fl) .or. isInf(fl) .or. 
     +         isNaN(fh) .or. isInf(fh) ) then 
           iflag = -2
           return 
           endif
 
         if ( fl .lt. 0.) then 
            xl = x1
            xh = x2
         else
            xh = x1
            xl = x2
         endif
                
         call funcd( rtsafe, f, df, e, a )
 
c        if for any of f, df, e  exception occured, exit
         if( isNaN(f) .or. isInf(f) .or. isNaN(df) .or. 
     +        isInf(df) .or. isNaN(e) ) then 
              iflag = -2
              return
           endif
         
         do j = 1, maxit,1
            
c          Condition calculation 
c          Use bisection whenever NR out of range or making smaller steps
 
c          NR Out of range condition  
c                ((rtsafe-xh)*df-f)*((rtsafe-xl)*df-f) .ge. 0.
c          NR Steps not big enough 
c                abs(2.*f) .gt. abs(dxold*df)
 
c          change this condition    
           
           outa = ((rtsafe-xh)*df-f)
           outb = ((rtsafe-xl)*df-f)
           
           big  = abs(2.*f) .gt. abs(dxold*df)
 
c          Check if any exception occured           
 
           if ( isNaN(big) .or. isNaN(outa) .or. isNaN(outb)
     +        .or. isInf(outa) .or. isInf(outb) ) then 
             iflag = -2 
             return
           endif
 
c          Do next step           
            
           if ( sgn(outa) * sgn(outb) .ge. 0. .or. big ) then 
 
c             Do bisection step instead of NR step           
              dxold = dx 
              dx = .5 * ( xh - xl )
              rtsafe = xl + dx
              
              if ( xl .eq. rtsafe ) then
c                root successfully found
                 iflag = 1
*                 print *, 'success 1'
		 return 
              endif
 
           else
 
c             Do NR step  
              dxold = dx
              dx = f/df
              temp = rtsafe
              rtsafe = rtsafe - dx
              
              if ( temp .eq. rtsafe ) then
c                root successfully found                   
                 iflag = 1
*		 print *, 'success 2'
                 return
              endif
 
           endif
c
c
c          Estimate roundoff error
           e = mepsd * e   
 
c          no overflow is harmful same for underflow
 
           if( abs(f) .le. e ) then
c             due to roundoff we can exit 
               iflag = 0
*	       print *, e, j
               return
 
           endif
c
           call funcd( rtsafe, f, df, e, a )
 
c          if for any of f, df, e  exception occured, exit
 
           if( isNaN(f) .or. isInf(f) .or. 
     +         isNaN(df) .or. isInf(df) .or. 
     +         isNaN(e) ) then 
              isflag = -2
              return
           endif
           
c          Maintain bracket 
           if( f .lt. 0. ) then
              xl = rtsafe
           else
              xh = rtsafe
           endif
c  
        enddo     
 
        iflag = -1
c       RTsafe exceeding max number of iterations
 
        return 
        end  
 
 
 
c**********************************************************************
c     POLYNOMIAL CALCULATION ( degree 3 )
c 
c     This function given point x and coefficients of qubic returns
c     p(x), derivative dp(x) and estimated roundoff error e.
c
c**********************************************************************
 
      subroutine funcd( x, p, dp, e, a )
         parameter(n=3)
         real a(0:3), x, p, dp, e
 
 
         p =  a(n)
         e =  .5 * abs(p)
         dp = 0.
 
         do i=n-1,0,-1
            dp = p + dp*x
            p = a(i) + p*x
            e = abs(x)*e + abs(p)
         enddo
         return
 
      end                                                   
 
c**********************************************************************
c     POLYNOMIAL CALCULATION
c 
c     This function given point x and coefficients a of polynomial 
c     of the degree 3 returns p(x).
c
c**********************************************************************
 
      real function poli( x, a )
         parameter( n=3)
         real a(0:3), x
         real p
 
         p =  a(n) 
         do i=n-1,0,-1
            p = a(i) + p * x
         enddo
         poli = p 
         return
      end          
                                         
c
c**********************************************************************
c Function  
c
c       x = 0 return 0
c       x < 0 return -1
c       x > 0 return 1       
c
        real function sgn( x )
           real x 
 
           if ( x .eq. 0. ) then 
              sgn = 0.
           else if ( x .lt. 0. ) then 
              sgn = -1. 
           else 
              sgn = 1. 
           endif 
        end
